/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#ifndef BEATTRACKER_H_
#define BEATTRACKER_H_

#include <vector>
#include <fcntl.h>
#include <string>
#include <iostream>
#include <fstream>
//#include <boost/multi_array.hpp>
//#include <boost/asio.hpp>
//#include <boost/lexical_cast.hpp>
#include "ring_i_buffer.hpp"
#include "fft_analysis.h"

//using namespace boost;
//using namespace boost::asio;
//using namespace boost::asio::ip;

namespace beat_tracker {
/**
 * @brief	ビートトラッキング結果データクラス
 */
class BeatTrackerData {
public:
    /**
     * @brief	初期化
     */
    BeatTrackerData() {
        interval_            = 0.0;
        beat_type_           = 0;
        current_time_        = 0.0;
        beat_time_           = 0.0;
        next_beat_time_      = 0.0;
        beat_diff_time_      = 0.0;
        next_beat_diff_time_ = 0.0;
        music_reliability_   = 0.0;
        
				//onsetVector_.assign(filter_num, 0);
				//STPMVector_.assign(cc_shift_diff, 0);
    };

		void setOnsetVector(std::vector<double> &onset)
		{
			onsetVector_ = onset;
		};
		void setSTPMVector(std::vector<double> &result)
		{
			STPMVector_ = result;
		};
		void setChromaVector(std::vector<double> &chroma)
		{
			chromaVector_ = chroma;
		};
		std::vector<double> getOnsetVector()
		{
			return onsetVector_;
		};
		std::vector<double> getSTPMVector()
		{
			return STPMVector_;
		};
		std::vector<double> getChromaVector()
		{
			return chromaVector_;
		};

    void setInterval(double interval) {
        interval_ = interval;
    };
    void setBeatType(size_t beat_type) {
        beat_type_ = beat_type;
    };
    void setBeatTimes(double beat_time,double current_time) {
        current_time_   = current_time;
        beat_time_      = beat_time;
        beat_diff_time_ = current_time-beat_time;
    };
    void setRecentBeatTime(double next_beat_time,double current_time) {
        next_beat_time_      = next_beat_time;
        next_beat_diff_time_ = current_time-next_beat_time;
    };
    void setBeatDiffTime(double beat_diff_time) {
        beat_diff_time_      = beat_diff_time;
    };
    void setMusicReliability(double music_reliability) {
        music_reliability_   = music_reliability;
    };

    double getInterval() {
        return interval_;
    };
    size_t getBeatType() {
        return beat_type_;
    };
    double getCurrentTime() {
        return current_time_;
    };
    double getBeatTime() {
        return beat_time_;
    };
    double getBeatDiffTime() {
        return beat_diff_time_;
    };
    double getRecentBeatTime() {
        return next_beat_time_;
    };
    double getRecentBeatDiffTime() {
        return next_beat_diff_time_;
    };
    double getMusicReliability() {
        return music_reliability_;
    };
    double calNextBeatTime(unsigned int current_time) {
        return 0.0;//TODO
    }

//		void sendmessage_ioi  (udp::socket& socket, double ioi, double next);
//		void sendmessage_start(udp::socket& socket);
//		void sendmessage_end  (udp::socket& socket);

		std::vector<double> onsetVector_;
		std::vector<double> STPMVector_;
		std::vector<double> chromaVector_;
private:
    double interval_;
    size_t beat_type_;
    double current_time_;
    double beat_time_;
    double next_beat_time_;
    double beat_diff_time_;
    double next_beat_diff_time_;
    double music_reliability_;
};
/**
 * @brief	ビートトラッキング初期化データクラス
 */
class BeatTrackerInit {
public:
    BeatTrackerInit() {
        sampling_freq_  = 44100.0;
        window_size_    = 4096;
        shift_size_     = 512;
        //メルフィルタバンク次元,周波数（メル次元）スタート,エンド
        filter_num_     = 64;
        mel_f_start_    = 1;
        mel_f_end_      = 63;
        //パターンマッチする幅, ずらす幅（スタート）,ずらす幅（おわり)
        cc_size_        = 1.01;//[sec]
        cc_shift_start_ = 0.51;//[sec]
        cc_shift_end_   = cc_size_;

        debug_flag_     = true;
        log_prefix_     = "";
    };
    ~BeatTrackerInit() {};
    void showAllSetting() {
        std::cout << "Sampling freq :¥t" << sampling_freq_ << "Hz" << std::endl;
    };
    // setter method
    void setInputSignal(double sampling_freq, size_t window_size, size_t shift_size) {
        sampling_freq_ = sampling_freq;
        window_size_   = window_size;
        shift_size_    = shift_size;
    };
    void setDebugFlag(bool debug_flag, std::string log_prefix) {
        debug_flag_    = debug_flag;
        log_prefix_    = log_prefix;
    };
    void setCCValues(double cc_size, double cc_shift_start, double cc_shift_end) {
        cc_size_        = cc_size;//1.01
        cc_shift_start_ = cc_shift_start;//0.51
        cc_shift_end_   = cc_shift_end;//1.01
    };
    // getter method
    double getSamplingFreq() {
        return sampling_freq_;
    };
    size_t getWindowSize() {
        return window_size_;
    };
    size_t getShiftSize() {
        return shift_size_;
    };
    bool getDebugFlag() {
        return debug_flag_;
    };
    std::string getLogPrefix() {
        return log_prefix_;
    };
    size_t getFilterNum() {
        return filter_num_;
    }
    size_t getMelFStart() {
        return mel_f_start_;
    }
    size_t getMelFEnd() {
        return mel_f_end_;
    }
    double getCcSize() {
        return cc_size_;
    }
    double getCcShiftStart() {
        return cc_shift_start_;
    }
    double getCcShiftEnd() {
        return cc_shift_end_;
    }
private:
    /**
     * 入力スペクトルの仕様
     */
    //入力信号のサンプリングレート
    double sampling_freq_;
    //FFT窓幅,SFTシフト幅
    size_t window_size_,shift_size_;

    /**
     * RealTimeBeatTracker パラメータ
     */
    //___________________メルフィルタバンク次元,周波数（メル次元）スタート,エンド
    size_t filter_num_,mel_f_start_,mel_f_end_;
    //___________________パターンマッチする幅, ずらす幅（スタート）,ずらす幅（おわり)
    double cc_size_,cc_shift_start_,cc_shift_end_;

    /**
     * ディバック用
     */
    //___________________デバックファイルを作るか
	bool debug_flag_;
	 //___________________デバックファイルの接頭語
	std::string log_prefix_;
};

/**
 * @brief	ビートトラッキング解析クラス
 */
class BeatTracker {
public:
    BeatTracker(BeatTrackerInit bt_init);
    virtual ~BeatTracker();

		void SetFreqBank();

    BeatTrackerData analyze(const std::vector<double>& new_spec);
    std::vector<double> MakeIncSpec(RingIBuffer<std::vector<double> >& spec);
    std::vector<double> MakeChromaVec(RingIBuffer<std::vector<double> >& spec);
    std::vector<double> MelFilter(RingIBuffer<std::vector<double> >& spec);
    std::vector<double> Filter3x3(RingIBuffer<std::vector<double> >& spec,int filter[][3]);
    std::vector<double> STPattanMatching(RingIBuffer<std::vector<double> >& input_signal);
		
private:
    double sampling_freq_; //サンプリング周波数
    size_t fft_freq_; //fft後の周波数数
    size_t shift_size_; //シフト長

		//___________________メルフィルタバンク次元
		int filter_num_;
		//___________________ずらす幅（スタート）
		int cc_shift;
		//___________________ずらす幅（おわり）
		int cc_shift_end;
		//___________________パターンマッチする幅
		int cc_size;
		//___________________周波数（メル次元）スタート
		int mel_f_start_;
		//___________________周波数（メル次元）エンド
		int mel_f_end_;

    bool debug_flag_;

    std::vector<double> melfilterbank_;//閾値列

    // ftごとにアップデート
    RingIBuffer<std::vector<double> > spec_;
    RingIBuffer<std::vector<double> > inc_spec_;
    RingIBuffer<std::vector<double> > mel_;
    RingIBuffer<std::vector<double> > inc_mel_;
    RingIBuffer<double> inc_mel_sum_;
    RingIBuffer<std::vector<double> > close_beats_reliability_;
    RingIBuffer<std::vector<double> > beats_reliability_mask_;
    RingIBuffer<double> beats_reliability_;
    RingIBuffer<size_t> beats_reliability_sum_num_;//今まで何回足しているか

    RingIBuffer<double> beat_intervals;//インターバル保存
    std::vector<double> beat_intervals_temp;

    RingIBuffer<double> music_reliability_;
    std::vector<double> music_reliability_temp;

		//chrome freq bank
		std::vector<double> vec_f_cent_bank_;
		std::vector<double> vec_f_cent_c_;
		std::vector<double> vec_f_cent_h_;
		std::vector<double> vec_bpf_ch_;
		
    // frequency analysis setting
    //boost::shared_ptr<WindowFunction> window_;
    //boost::shared_ptr<FftAnalysis> fft_;
#if EXTERNAL_FFT == 0
    FftAnalysis fft_;
#endif

    // output
    int new_frames_;



    //SemiBlind* psemiblind_;

    std::string log_prefix_;
    // file output
    std::ofstream ofs_spec;						//スペクトル
    std::ofstream ofs_inc_spec;					//スペクトル微分
    std::ofstream ofs_melfilterbank;			//メルスペクトル
//	std::ofstream ofs_mfcc;			//メルスペクトル
    std::ofstream ofs_chroma_vector;			//クロマベクトル
    std::ofstream ofs_inc_mel;					//メルスペクトル微分
    std::ofstream ofs_interval_value;			//
    std::ofstream ofs_interval_long_value;		//相互相関関数
    std::ofstream ofs_interval_contrast;
    std::ofstream ofs_interval_long_contrast;
    std::ofstream ofs_beats_reliability_mask;
    std::ofstream ofs_beats_reliability;
    std::ofstream ofs_beats_times;

//	std::ofstream ofs_inp_spec;
//	std::ofstream ofs_hum_spec;
//	std::ofstream ofs_can_spec;
//
    std::ofstream ofs_flat_num;
    std::ofstream ofs_music_reli;

    /*
     * analyze用作業変数
     */
    int t_current;
    int beat_time;
    int beat_type;
    int beat_interval;
    double music_reliability;
    bool music_signaloff_flag;
};



}

#endif /* BEATTRACKER_H_ */
