/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#define NDEBUG

#include <iostream>
#include <memory>
#include <list>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <math.h>
#include <Source.h>
#include <string>
#include <stdlib.h>
#include <stdio.h>

#include <typeinfo>

#include "BeatTracker.hpp"

#include <vector>
#include <beattracker/RealTimeBeatTrackerInit.hpp>
#include <beattracker/beat.h>
//#include <boost/asio.hpp>
//#include <beattracker/window_function.h>
#include <beattracker/fft_analysis.h>

#include <beattracker/beattracker.h>
#include <beattracker/fd_ica.h>
#include <beattracker/voice_cancel.h>
#include <complex>

using namespace std;
//using namespace boost;
using namespace FD;
//using namespace boost::numeric;
//using namespace boost::asio;
//using namespace boost::asio::ip;

//#define PORT_ROBOT 21007 //サーバープログラムとポート番号を合わせてください
//#define IPADDRESS  "192.168.19.100"

class BeatTracker;

DECLARE_NODE(BeatTracker);
/*Node
 *
 * @name BeatTracker
 * @category HARK:MUSIC
 * @description Beat tracking module
 *
 * @input_name AUDIO
 * @input_type Matrix<float>
 * @input_description Windowed sound data.
 *
 * @input_name FFTAUDIO
 * @input_type Matrix<complex<float> >
 * @input_description Fourier transformed signal. BeatTracker uses only channel-0.
 *
 * @output_name BEAT_DETECTED
 * @output_type bool
 * @output_description  Is a beat detected?
 *
 * @output_name TEMPO
 * @output_type float
 * @output_description beat tempo
 *
 * @output_name ONSET_VECTOR
 * @output_type Vector<float>
 * @output_description Mel-scaled Spectrogram after Sobel filtering
 *
 * @output_name STPM_VECTOR
 * @output_type Vector<float>
 * @output_description The result of STPM
 *
 * @output_name CHROMA_VECTOR
 * @output_type Vector<float>
 * @output_description The chroma vector of input signal
 *
 * @output_name CONFIDENCE
 * @output_type float
 * @output_description The confidence about the output beat.
 *
 * @parameter_name RELIABILITY-THRESHOLD
 * @parameter_type float
 * @parameter_value 0.1
 * @parameter_description The threshold for the beat detection.
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 44100
 * @parameter_description sampling rate of sound data.
 *
 * @parameter_name MAX_BPM
 * @parameter_type int
 * @parameter_value 120
 * @parameter_description Maximum BPM. 
 *
 * @parameter_name MIN_BPM
 * @parameter_type int
 * @parameter_value 60
 * @parameter_description Minimum BPM. 
 *
 * @parameter_name VECTOR_FLAG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Output vectors or not. 
 *
 * @parameter_name LOG_FLAG
 * @parameter_type bool
 * @parameter_value false
 * @parameter_description Log or not. 
END*/

#if 1

#define DBG(x)

#define DBGPRINT(x) DBG(std::cout << #x << " = " << x << endl;)

#define FILTER_NUM   64
#define FILTER_START 1
#define FILTER_END   63

void PrintVector(std::vector<double> &vect, int count = 5)
{
    for (int i = 0; i < count && i < vect.size(); i++)
    {
        std::cout << vect[i] << ", ";
    }
    std::cout << std::endl;
}

void update_signal
(
    std::vector<double>& signal,
    Matrix<float>& src,
    int updated_samples
)
{
    const int samples = static_cast<int>(signal.size());
    const int reserved_samples = samples - updated_samples;
    int channels = src.nrows();

    // extract old data
    //  updated: [0, ..., updated_samples - 1]
    // reserved: [updated_samples, ..., samples - 1]
    vector<double> tmp_signal(samples);
    copy(signal.begin() + updated_samples,
         signal.begin() + samples, tmp_signal.begin());

    double sig;
    for (int j = 0; j < updated_samples; j++)
    {
        sig = 0;
        /*
         * チャンネルの総和
         * オリジナルソースにおいて、LチャンネルとRチャンネルの和を取っているのと対応する。
         */
        for (int i = 0; i < channels; i++)
        {
            // ここが8フレーム遅れる原因？
            sig += src(i, reserved_samples + j) * 0.5;
        }
        //tmp_signal[reserved_samples + j] = (sig / channels) / 32768.0;
        tmp_signal[reserved_samples + j] = sig;
    }

    // update data
    // reserved: [0, ..., reserved_samples - 1]
    //  updated: [reserved_samples, ..., samples - 1]
    signal.swap(tmp_signal);
}

void update_fftsignal
(
    std::vector<FftComplex>& signal,
    Matrix<std::complex<float> >& src,
    int updated_samples
)
{
    const int samples = static_cast<int>(signal.size());
    const int reserved_samples = samples - updated_samples;
    int channels = src.nrows();

    // extract old data
    //  updated: [0, ..., updated_samples - 1]
    // reserved: [updated_samples, ..., samples - 1]
    vector<FftComplex> tmp_signal(samples);
    copy(signal.begin() + updated_samples,
         signal.begin() + samples, tmp_signal.begin());

    std::complex<float> c;
    double sig;
    for (int j = 0; j < updated_samples; j++)
    {
        c = src(0, reserved_samples + j);
        tmp_signal[reserved_samples + j] = FftComplex(c.real(), c.imag());
    }

    // update data
    // reserved: [0, ..., reserved_samples - 1]
    //  updated: [reserved_samples, ..., samples - 1]
    signal.swap(tmp_signal);
}

#endif

class BeatTracker : public BufferedNode {
    int audioInputID;
    int fftAudioInputID;
    int beatDetectedOutputID;
    int tempoOutputID;
    int confidenceOutputID;

    int onsetVectorID;
    int STPMVectorID;
    int chromaVectorID;
    
    float reliabilityThreshold;
    int advance;
    int samplingRate;

    int maxBPM;
    int minBPM;
  	bool vector_flag;
  	bool log_flag;

    // std::vector<FftComplex> fft_voice_spec, fft_mic_spec;

    // Signal setting
    std::shared_ptr<std::vector<double> > mic_signal_ptr;
    std::shared_ptr<std::vector<double> > voice_signal_ptr;
    std::shared_ptr<std::vector<double> > mic_spec_ptr;
    std::shared_ptr<std::vector<double> > voice_spec_ptr;

    std::vector<FftComplex> fft_voice_spec, fft_mic_spec;
    std::shared_ptr<std::vector<FftComplex> > fft_mic_spec_ptr;

    //VoiceCancel
    std::shared_ptr<VoiceCancel> voice_cancel_ptr;

    //前のビート時刻
    int old_beat_time;
    //前に取得した時刻
    unsigned int last_time;
    //ビートトラッキング結果収納用変数
    beat_tracker::BeatTrackerData result;

    std::list<int> mlFutureBeatTime;
    int mnLastFutureBeatTime;
    bool mbFutureBeatTimeSet;
    int mnLastBeatDetectedTime;
    int mnLastAppendedBeatTime;

    ////////////////////////////
    //BeatTrackerで用いる変数
    ////////////////////////////
    //入力スペクトルの仕様
    double sampling_freq_;//入力信号のサンプリングレート
    size_t window_size_,shift_size_;//FFT窓幅,SFTシフト幅

    bool goto_method_flag_;
    bool voice_cancel_flag_;
    std::string prf_file_name_;

    SemiBlind *psemiblind_;

    beat_tracker::BeatTracker* beat_track_;
    WindowFunction* window_;
#if EXTERNAL_FFT == 0
    FftAnalysis* fft_;
#endif
    Beat* detected_beats_;

	  //for ip connecting
//	io_service io_service_robot;
//	string     str_ipaddress_robot;//"192.168.19.100");//IPADDRESS);
//	int        port_robot;
//	udp::socket socket_robot;//(io_service_robot);
//	boost::system::error_code error_robot;
	ofstream   output_muratabt;
	
public:
    BeatTracker(string nodeName, ParameterSet params)
        : BufferedNode(nodeName, params),
          detected_beats_(0),
          beat_track_(0), window_(0),
#if EXTERNAL_FFT == 0
          fft_(0),
#endif
          old_beat_time(0), last_time(0)//,
//					socket_robot(io_service_robot)
    {
        std::cout << "BeatTracker::BeatTracker" << endl;

        audioInputID         = addInput("AUDIO");
        fftAudioInputID      = addInput("FFTAUDIO");
        beatDetectedOutputID = addOutput("BEAT_DETECTED");
        tempoOutputID        = addOutput("TEMPO");
        confidenceOutputID   = addOutput("CONFIDENCE");
        onsetVectorID        = addOutput("ONSET_VECTOR");
        STPMVectorID         = addOutput("STPM_VECTOR");
        chromaVectorID       = addOutput("CHROMA_VECTOR");

        ObjectRef param      = parameters.get("RELIABILITY-THRESHOLD");
        reliabilityThreshold = dereference_cast<float>(param);

        param        = parameters.get("ADVANCE");
        advance      = dereference_cast<int>(param);

        param        = parameters.get("SAMPLING_RATE");
        samplingRate = dereference_cast<int>(param);
        
        param        = parameters.get("MAX_BPM");
        maxBPM       = dereference_cast<int>(param);
        
        param        = parameters.get("MIN_BPM");
        minBPM       = dereference_cast<int>(param);

        param        = parameters.get("VECTOR_FLAG");
        vector_flag  = dereference_cast<bool>(param);

        param        = parameters.get("LOG_FLAG");
        log_flag     = dereference_cast<bool>(param);

        inOrder                = true;
        mbFutureBeatTimeSet    = false;
        mnLastBeatDetectedTime = 0;
        mnLastAppendedBeatTime = 0;  

				output_muratabt.open("bt_muratabt.dat");
    }

    // Function by olivier
    /*void reset()
    {
      RealTimeBeatTrackerInit init_data;

      // ビートトラッキングの設定値を格納
      init_data.set_sampling_freq(samplingRate);
      init_data.set_beat_tracking_method(STPM_METHOD);
      init_data.SetSTFTParameter(audioInput.ncols(), this->advance);
      init_data.SetMelFilterParameter(64,1,63);
      init_data.SetSTPMParameter(1.01,0.51,1.01);
      init_data.SetDebugParameter(false,true,"");

      this->realTimeBeatTrackerInitialize(init_data);
    }*/

    virtual ~BeatTracker()
    {
        std::cout << "BeatTracker::~BeatTracker" << endl;

        delete this->detected_beats_;
        delete this->beat_track_;
        delete this->window_;
#if EXTERNAL_FFT == 0
        delete this->fft_;
#endif
    }

  virtual void initialize()
  {
    outputs[beatDetectedOutputID].lookAhead
      = outputs[tempoOutputID].lookAhead
      = outputs[confidenceOutputID].lookAhead
      = outputs[onsetVectorID].lookAhead
      = outputs[STPMVectorID].lookAhead
      = outputs[chromaVectorID].lookAhead
      = 1 + 
      	max(
		  	max(
		  		max(
		  			max(
		  				max(
		  					outputs[beatDetectedOutputID].lookAhead,
							outputs[tempoOutputID].lookAhead),
						outputs[confidenceOutputID].lookAhead),
					outputs[onsetVectorID].lookAhead),
				outputs[STPMVectorID].lookAhead),
			outputs[chromaVectorID].lookAhead);

    outputs[beatDetectedOutputID].lookBack
      = outputs[tempoOutputID].lookBack
      = outputs[confidenceOutputID].lookBack
      = outputs[onsetVectorID].lookBack
      = outputs[STPMVectorID].lookBack
      = outputs[chromaVectorID].lookBack
      = 1 + 
      	max(
		  	max(
		  		max(
		  			max(
		  				max(
		  					outputs[beatDetectedOutputID].lookBack,
							outputs[tempoOutputID].lookBack),
						outputs[confidenceOutputID].lookBack),
					outputs[onsetVectorID].lookBack),
				outputs[STPMVectorID].lookBack),
			outputs[chromaVectorID].lookBack);

    this->BufferedNode::initialize();
  }

    void count0Initialize(int output_id, Buffer &out)
    {
        std::cout << "BeatTracker::count0Initialize" << endl;

        RCPtr<Matrix<float> > audioInputPtr = this->getInput(this->audioInputID, 0);
        Matrix<float>& audioInput = *audioInputPtr;

        RealTimeBeatTrackerInit init_data;

        // ビートトラッキングの設定値を格納
        init_data.set_sampling_freq(samplingRate);
        init_data.set_beat_tracking_method(STPM_METHOD);
        init_data.SetSTFTParameter(audioInput.ncols(), this->advance);
        init_data.SetMelFilterParameter(FILTER_NUM,FILTER_START,FILTER_END);
        init_data.SetDebugParameter(false,log_flag,"");//center->debugpara
        //maybe don't used following parameters. See line 40 in audioPreprocessMethod.cpp
        init_data.SetSTPMParameter(minBPM,
                                   maxBPM,
                                   minBPM);//1.51,0.76,1.51);

        this->detected_beats_ = new Beat(2,0.0,0.0,0.0, samplingRate);

				//for ip connection
//				str_ipaddress_robot = IPADDRESS;
//				port_robot          = PORT_ROBOT;
//				cout << "ROBOT  IP : " <<  str_ipaddress_robot << ":" << port_robot << endl;

        this->realTimeBeatTrackerInitialize(init_data);
    }

    void realTimeBeatTrackerInitialize(const RealTimeBeatTrackerInit& init_data)
    {
        std::cout << "BeatTracker::realTimeBeatTrackerInitialize" << endl;

        ////////////////////////////
        // 入力データ仕様設定
        ////////////////////////////
        sampling_freq_ = init_data.sampling_freq_;
        window_size_   = init_data.window_size_;
        shift_size_    = init_data.shift_size_;

        DBG(std::cout << "sampling:" << sampling_freq_ 
                      << ", window:" << window_size_ 
                      << ", shift:"  << shift_size_ << endl;)

        ////////////////////////////
        // ディバック設定
        ////////////////////////////
        voice_cancel_flag_ = init_data.voice_cancel_flag_;
        prf_file_name_     = init_data.log_prefix_name_;

				////////////////////////////
				// ビートトラッキング手法選択
				////////////////////////////
        if(init_data.beat_tracking_method_ == SELF_CORRELATION_METHOD) {
            goto_method_flag_ = true;
        } else {
            goto_method_flag_ = false;
        }

#if EXTERNAL_FFT == 0
        ////////////////////////////
        // ＦＦＴオブジェクト初期化
        ////////////////////////////
        window_ = new WindowFunction(window_size_,WindowFunction::WIN_HAMMING);
        fft_    = new FftAnalysis(*window_);
#endif

				////////////////////////////
				// ビートトラッキングオブジェクト初期化
				////////////////////////////
        beat_tracker::BeatTrackerInit init_beat_tracker_data;
        init_beat_tracker_data.setInputSignal(init_data.sampling_freq_, init_data.window_size_, init_data.shift_size_);
        init_beat_tracker_data.setDebugFlag(init_data.debug_flag_,init_data.log_prefix_name_);
				init_beat_tracker_data.setCCValues(init_data.cc_size_, init_data.cc_shift_start_, init_data.cc_shift_end_);
        beat_track_ = new beat_tracker::BeatTracker(init_beat_tracker_data);

        // Signal setting
        this->mic_signal_ptr   = std::shared_ptr<std::vector<double> >(new std::vector<double>(window_size_, 0.0));
        this->voice_signal_ptr = std::shared_ptr<std::vector<double> >(new std::vector<double>(window_size_, 0.0));
        this->mic_spec_ptr     = std::shared_ptr<std::vector<double> >(new std::vector<double>(window_size_ / 2, 0.0));
        this->voice_spec_ptr   = std::shared_ptr<std::vector<double> >(new std::vector<double>(window_size_ / 2, 0.0));

        //VoiceCancel
        this->voice_cancel_ptr = std::shared_ptr<VoiceCancel>(new VoiceCancel(window_size_));

        this->fft_mic_spec_ptr = std::shared_ptr<std::vector<FftComplex> >(new std::vector<FftComplex>(window_size_ / 2 + 1, 0.0));
    }


    void calculate(int output_id, int count, Buffer &out)
    {
			//cout << "test1" << endl;
        RCPtr<Matrix<float> > audioInputPtr = this->getInput(this->audioInputID, count);
        Matrix<float>& audioInput = *audioInputPtr;

        RCPtr<Matrix<complex<float> > > fftAudioInputPtr = this->getInput(this->fftAudioInputID, count);
        Matrix<complex<float> >&        fftAudioInput    = *fftAudioInputPtr;
    
        if (count == 0)
        {
            this->count0Initialize(output_id, out);

            // 初めは空なので初期化する
#if EXTERNAL_FFT == 0
            update_signal(*this->mic_signal_ptr, audioInput, static_cast<int>(this->mic_signal_ptr->size() /*window_size_*/));
#else
            update_signal(*this->mic_signal_ptr, audioInput, static_cast<int>(this->mic_signal_ptr->size() /*window_size_*/));
            update_fftsignal(*this->fft_mic_spec_ptr, fftAudioInput, static_cast<int>(fftAudioInput.ncols()));
#endif
        }
        else
        {
#if EXTERNAL_FFT == 0
            update_signal(*this->mic_signal_ptr, audioInput, this->advance);
#else
            update_signal(*this->mic_signal_ptr, audioInput, this->advance);
            //update_fftsignal(*this->fft_mic_spec_ptr, fftAudioInput, this->advance / 2);
            update_fftsignal(*this->fft_mic_spec_ptr, fftAudioInput, static_cast<int>(fftAudioInput.ncols()));
#endif
        }

        DBG(std::cout << "mic_signal_ptr:";)
        DBG(PrintVector(*this->mic_signal_ptr);)

#if EXTERNAL_FFT == 0
        // スペクトルへ変換
        fft_->execute_fourier_transform(mic_signal_ptr->begin(), fft_mic_spec);
#else
        /*
        int i;

        for (int i = 0; i < mic_signal_ptr->size(); i++)
        {
          std::cout << (*mic_signal_ptr)[i] << " ";
        }
        std::cout << std::endl;

        fft_->execute_fourier_transform(mic_signal_ptr->begin(), fft_mic_spec);

        for (i = 0; i < fft_mic_spec.size(); i++)
        {
          if (fft_mic_spec[i].abs() != (*fft_mic_spec_ptr)[i].abs())
          {
            std::cout << "external fft" << (*fft_mic_spec_ptr)[i].abs() << " != internal fft" << fft_mic_spec[i].abs() << "   at " << i << "  ext - int = " << (*fft_mic_spec_ptr)[i].abs() - fft_mic_spec[i].abs() << std::endl;
            break;
          }
        }
        if (i == fft_mic_spec.size())
        {
          std::cout << "external fft == internal fft" << std::endl;
        }
        */
#endif

        // ヴォイスキャンセル実行
        if (voice_cancel_flag_)
        {
            /*
             * ボイスキャンセル機能は実装しない
             */
#if 0
            fft_->execute_fourier_transform(voice_signal_ptr->begin(),
                                            fft_voice_spec);
            voice_cancel_ptr->ExecuteVoiceCancel(fft_voice_spec, fft_mic_spec,
                                                 *mic_spec_ptr);
#endif
        } else
        {
#if EXTERNAL_FFT == 0
            for (size_t f = 0; f < window_size_ / 2; f++)
                (*mic_spec_ptr)[f] = fft_mic_spec[f].abs();
#else
            for (size_t f = 0; f < window_size_ / 2; f++)
                (*mic_spec_ptr)[f] = (*fft_mic_spec_ptr)[f].abs();
#endif
        }

        //ビートトラッキングし，結果を収納
        result = beat_track_->analyze(*mic_spec_ptr);

        //ビート絶対時刻を取得
        int beat_time            = result.getBeatTime();
        int beat_interval        = result.getInterval();
        int current_time         = result.getCurrentTime();
        double music_reliability = result.getMusicReliability();

        //cout << "Current time: " << current_time << endl;

        double beat_tempo = 0;
        bool beat_detected = false;
        //bool beat_detected = old_beat_time < current_time; // line by olivier

        int nFutureBeatTime = result.getRecentBeatTime() + beat_interval*2;

        //cout << "Future beat time: " << nFutureBeatTime << endl;

        if( nFutureBeatTime > mnLastAppendedBeatTime + beat_interval*0.8 )
        {
            //cout << "Appended beat time: " << nFutureBeatTime << " Last appended: " << mnLastAppendedBeatTime << " Beat interval: " << beat_interval << endl;
            mlFutureBeatTime.push_back( nFutureBeatTime );
            mnLastAppendedBeatTime = nFutureBeatTime;
        }

				if( mlFutureBeatTime.size() > 0 && mlFutureBeatTime.front() <= current_time)
				{
            //cout << "Popped beat with time " << mlFutureBeatTime.front() << " at time " << current_time << endl;
					mlFutureBeatTime.pop_front();
					beat_detected = true;
            //cout << "Beat buffer length: " << mlFutureBeatTime.size() << endl;
            
					if(music_reliability >= reliabilityThreshold)
					{
						std::cout << "Beat detected at time: " << current_time << endl;
						
						beat_tempo = 60.0 / ((beat_interval * shift_size_) / this->sampling_freq_);
						detected_beats_->set_beat(beat_time     * shift_size_ / sampling_freq_,
						                          beat_interval * shift_size_ / sampling_freq_,
						                          music_reliability);
						cout << "Rely!!! : " << music_reliability << " tempo: "<< beat_tempo <<endl;
						// murata bt
						output_muratabt << beat_tempo << " " << current_time << " " << music_reliability << endl;

						double beat_interval_msec = beat_interval * shift_size_ / sampling_freq_ * 1000;
						//result.sendmessage_ioi(socket_robot, beat_interval_msec, beat_interval_msec);
						// cout << "send message : " 
						// 		 << beat_interval_msec << ", " 
						// 		 << beat_interval_msec << endl;;

					}
					else
					{
						cout << "Discard : " << music_reliability << endl;
					}
        }//*/

				(*outputs[this->beatDetectedOutputID].buffer)[count] 
					  = beat_detected ? TrueObject : FalseObject;
        (*outputs[this->confidenceOutputID].buffer)[count]
            = ObjectRef(Float::alloc((float)music_reliability));
        (*outputs[this->tempoOutputID].buffer)[count]
            //= ObjectRef(Float::alloc((float)beat_tempo));
            = beat_detected ? ObjectRef(Float::alloc((float)beat_tempo)) 
                            : ObjectRef(Float::alloc(0.0));
				// /*
				if(vector_flag)
				{
					RCPtr<Vector<float> > onsetVectorPtr (Vector<float>::alloc(0));
					RCPtr<Vector<float> > STPMVectorPtr  (Vector<float>::alloc(0));
					RCPtr<Vector<float> > chromaVectorPtr(Vector<float>::alloc(0)); // chroma
					Vector<float>& onsetVectorData  = *onsetVectorPtr;
					Vector<float>& STPMVectorData   = *STPMVectorPtr;
					Vector<float>& chromaVectorData = *chromaVectorPtr; // chroma
					onsetVectorData.resize(FILTER_NUM);
					STPMVectorData.resize(result.STPMVector_.size());
					chromaVectorData.resize(12);
					
					//copy( result.onsetVector_ .begin(), result.onsetVector_ .end(), 
					//			back_inserter( onsetVectorData ));
					copy( result.onsetVector_ .begin(), result.onsetVector_ .end(), onsetVectorData .begin());
					copy( result.STPMVector_  .begin(), result.STPMVector_  .end(), STPMVectorData  .begin());
					copy( result.chromaVector_.begin(), result.chromaVector_.end(), chromaVectorData.begin());// chroma
					
					(*outputs[this->onsetVectorID ].buffer)[count] = onsetVectorPtr;
					(*outputs[this->STPMVectorID  ].buffer)[count] = STPMVectorPtr;
					(*outputs[this->chromaVectorID].buffer)[count] = chromaVectorPtr; // chroma
				}
				// */

				//cout << "test2" << endl;
    }

    IN_ORDER_NODE_SPEEDUP(BeatTracker);
};
